@@ -145,7 +145,7 @@ if bUseNumba :
                     xr = np.dot( Z.T,Vt )
                     return ( xr )
                 else:
-                    xr = np.tensordot( Z.T,Vt, axes=(0,1) )
+                    xr = np.dot( Z.T,Vt )
                     return ( xr.T )
 else :
         def distance_matrix_to_absolute_coordinates ( D , bSquared = False, n_dimensions=2 , bLegacy=True ):
@@ -577,7 +577,7 @@ def connectivity_clustering( distm:np.array, cutoff:float, bBool:bool=False ) ->
     else :
         res = connectivity( distm , cutoff )
     return ( res )
-        
+
 
 clustering_algorithm = None
 clustering_algorithm = sc.KMeans(10) # CHOOSE SOMETHING YOU LIKE NOT THIS
@@ -1080,7 +1080,10 @@ def sclinkages ( distm ,command='min' , bStrKeys=True ) -> dict :
     from scipy.spatial.distance  import squareform
     from scipy.cluster.hierarchy import linkage as sclinks
     from scipy.cluster.hierarchy import fcluster
-    Z  = sclinks( squareform(distm) , {'min':'single','max':'complete'}[command] )
+    cmd = command
+    if command in set(['min','max']) :
+        cmd = {'min':'single','max':'complete'}[command]
+    Z  = sclinks( squareform(distm) , cmd )
     CL = {}
     F  = {} # NEW
     for d in Z[:,2] :
@@ -1187,6 +1190,141 @@ def linkages ( distm:np.array , command:str='min' ,
         linkages_ = L
     return ( linkages_ )
 
+def create_cluster_lookups ( cluster_ids:list[int] ) -> dict :
+    pid2cid = {}
+    cid2pid = {}
+    for pid,cid in zip ( range(len(cluster_ids)) , cluster_ids ) :
+        pid2cid[pid] = cid
+        if cid in cid2pid :
+            cid2pid[cid] .append(pid)
+        else :
+            cid2pid[cid] = [pid]
+    return ( {'c2p':cid2pid ,'p2c':pid2cid } )
+
+def pair_compactness (  distm:np.array , bSelected:list[bool] = None,
+                        cluster1_indices:list[int] = None ,
+                        cluster2_indices:list[int] = None  ) -> list[float] :
+    if not bSelected is None :
+        bNot = lambda x: [ not x_ for x_ in x ]
+        g1   = np.where(bSelected)[0]
+        g2   = np.where(bNot(bSelected))[0]
+    else :
+        g1   = cluster1_indices
+        g2   = cluster2_indices
+    #
+    # BELOW SELECTION MUST BE FASTER
+    allPairs = lambda A, v1, v2 : np.array([ A[i,j] for i in v1 for j in v2 ]) # THIS IS AN ORDER OF MAGNITUDE FASTER THAN BELOW
+    # allPairs = lambda A, v1, v2 : np.array([ A[(p,q)] for (p,q) in zip(*np.meshgrid(v1,v2)) ])
+    #
+    if len(g2) == 0 :
+        g2 = g1
+    if len(g1) == 0 :
+        g1 = g2
+    #
+    pS		= np.sum( allPairs(distm,g1,g1) )
+    qS		= np.sum( allPairs(distm,g2,g2) )
+    pqS		= np.sum( allPairs(distm,g1,g2) )
+    FS          = np.sum( allPairs(distm,[*g1,*g2],[*g1,*g2]) )
+    p1122	= 1/len(g1)**2 * pS + 1/len(g2)**2 * qS
+    p1212	= 2*1/len(g1) * 1/len(g2) * pqS # p2121 = p1212
+    score	= p1212 - p1122
+    norm_scores = np.array([pS,qS,pqS])/FS
+    norm_scores = np.append(norm_scores, norm_scores[0] + norm_scores[1] - norm_scores[2] )
+    #
+    #	"QUADRATURE"	"TORQUE"	"NORMED DIFF"
+    return ( [ p1212 - p1122 , pqS/len(g1)/len(g2) - (qS/len(g1)/2 + pS/len(g2)/2 )*0.5 , norm_scores ] )
+
+
+def complete_compactness_score ( distm:np.array , cluster_ids:np.array ) :
+    lucids = list(set(cluster_ids))
+    lookup = { j:i for i,j in zip( range(len(lucids)),lucids ) }	# DID THE USER USE STRANGE INTS?
+    res_l  = create_cluster_lookups( cluster_ids )			# YES WE NEED BOTH
+    Z      = np.zeros( len(lucids)*len(lucids) ).reshape( len(lucids),len(lucids) )
+    Y	   = np.zeros( len(lucids)*len(lucids) ).reshape( len(lucids),len(lucids) )
+    X	   = np.zeros( len(lucids)*len(lucids) ).reshape( len(lucids),len(lucids) )
+    density_functional = lambda x:np.mean(x)
+    #
+    for cid1 in lucids :
+        for cid2 in lucids :
+            xid1 = lookup[cid1]
+            xid2 = lookup[cid2]
+            if xid1 < xid2:
+                cluster1_ids = res_l['c2p'][cid1]
+                cluster2_ids = res_l['c2p'][cid2]
+                res = pair_compactness ( distm , cluster1_indices=cluster1_ids , # BOTTLENECK
+                                                 cluster2_indices=cluster2_ids ) # THIS CALCULATION IS SYMMETRIC
+                score = res[0]
+                erocs = res[1]
+                nosco = res[2][-1]
+            elif xid1 > xid2 :
+                score = Z[xid2,xid1]
+                erocs = Y[xid2,xid1]
+                nosco = X[xid2,xid1]
+            else :
+                score = 0
+                erocs = 0
+                nosco = 0
+            Z [ xid1 , xid2 ] = score
+            Y [ xid1 , xid2 ] = erocs
+            X [ xid1 , xid2 ] = nosco
+    return ( [ Z , Y , X ] )
+
+def split_or_merge ( distm:np.array, cluster_ids:np.array ) -> dict :
+    density_functional = lambda x:np.mean(x)
+    res = complete_compactness_score ( distm , cluster_ids ) # BOTTLENECK
+    Z   = res[0]
+    density_cutoff = density_functional ( [ Z[i,j] for i in range(len(Z)) for j in range(len(Z)) if not i==j ] )
+    merge = Z[0]   <  density_cutoff
+    split = Z[0]   >= density_cutoff
+    keep  = res[2][-1] <= 0
+    desc_ ="""
+    compactness, torque, normed scores	- all information retained
+    keep			- strong suggestion
+    """
+    return ( { 'compactness':Z , 'torque':res[1] ,'normed scores':res[2][-1] , 'keep split':keep , 'desc':desc_ } )
+
+
+def cluster_appraisal( x:pd.Series , garbage_n = 0 , Sfunc = lambda x:np.mean(x,0)) -> list :
+    """
+Clustering Optimisation Method for Highly Connected Data
+Richard Tjörnhammar
+https://arxiv.org/abs/2208.04720v2
+    """
+    from collections import Counter
+    decomposition = [ tuple((item[1],item[0])) for item in Counter(x.values.tolist()).items() if item[1] > garbage_n ]
+    N     = len ( x )
+    A , B = 1 , N
+    if len(decomposition) > 0 :
+        A = Sfunc(decomposition)[0]
+        B = len(decomposition)
+    else :
+        decomposition = [ tuple( (0,0) ) ]
+    decomposition .append( tuple( ( B ,-garbage_n )) )
+    decomposition .append( tuple( ( A*B/(A+B) , None ) ) )
+    return ( decomposition )
+
+def generate_clustering_labels ( distm:np.array , cmd:str='min' , labels:list[str]=None ,
+                                 bExtreme:bool=False , n_clusters:int = None) -> tuple :
+    clabels_n , clabels_o = None , None
+    res         = sclinkages( distm , cmd )['F']
+    index       = list( res.keys() )
+    if labels is None :
+        labels = range(len(distm))
+    hierarch_df = pd.DataFrame ( res.values() , index=list(res.keys()) , columns = labels )
+    cluster_df  = hierarch_df .T .apply( lambda x: cluster_appraisal(x,garbage_n = 0) )
+    clabels_o , clabels_n = None , None
+    screening    = np.array( [ v[-1][0] for v in cluster_df.values ] )
+    level_values = np.array( list(res.keys()) )
+    if bExtreme :
+        imax            = np.argmax( screening )
+        clabels_o       = hierarch_df.iloc[imax,:].values.tolist()
+    if not n_clusters is None :
+        jhit = np.argmin([ np.abs(len(cluster_df.iloc[i])-2-n_clusters)\
+                   for i in range(len(cluster_df)) ])
+        clabels_n = hierarch_df.iloc[jhit,:].values.tolist()
+        print ( len(set(clabels_n)) , jhit ) ; exit(1)
+    return ( clabels_n , clabels_o , hierarch_df , np.array( [level_values,screening] ) )
+
 
 if __name__ == '__main__' :
 
@@ -1240,6 +1378,25 @@ if __name__ == '__main__' :
         T.append( time.time() )
         print ( len(R) , len( P[0] ) , len(Q) , np.diff(T) , alpha )
 
+
+    coordinates = np.array([ *np.random.randn(500,100) , *(np.random.randn(500,100) + 10) ])
+    print ( coordinates )
+    dm = distance_calculation( coordinates ,
+        distance_type = 'correlation,spearman' ,
+        #bRemoveCurse  = True , nRound=4 )
+        bRemoveCurse = False )
+    print ( dm )
+    print ( np.sum(np.diag(dm)) )
+    cln,clo,hdf,sol = generate_clustering_labels( dm , cmd='max' , bExtreme=True , n_clusters=80 )
+    imax = np.argmax( sol[1] )
+    print ( clo )
+    print ( sol[0] )
+    print ( sol[1] )
+    import matplotlib.pyplot as plt
+    plt.plot(sol[0],sol[1],'k')
+    plt.show()
+
+
     """    CONNECT +
  N  M  K   EDNESS LINEAR    IVITY JIT     EDNESS-JIT
 67 67 67 [7.61270523e-04 5.55515289e-05 2.30550766e-04] 1.01
