@@ -1800,11 +1800,36 @@ def correlation_distance ( agg_df:pd.DataFrame  , decimal_power:int = 10000 ,
             corr = spearmanrho(agg_df.values,agg_df.values)
 
         distm = pd.DataFrame( np.sqrt( 1 - correlation_transform(corr) ) ,
-                                    index   = agg_df.columns.values ,
-                                    columns = agg_df.columns.values )
+                                    index   = agg_df.index.values ,
+                                    columns = agg_df.index.values )
         distm = distm.apply(lambda x: np.round(x*decimal_power)/decimal_power )
     return ( distm )
 
+
+def distance_calculation ( coordinates:np.array ,
+                           distance_type:str , bRemoveCurse:bool=False ,
+                           nRound:int = None ) -> np.array :
+    crds = coordinates
+    if 'correlation' in distance_type :
+        if 'pearson' in distance_type :
+            corr =  pearsonrho( crds , crds )
+        else :
+            corr = spearmanrho( crds , crds )
+        corr = 0.5 * ( corr + corr.T ) # THIS SHOULD BE REMOVED ONCE DISC. LOCATED
+        if 'absolute' in distance_type :
+            corr = np.abs( corr )
+        if 'square' in distance_type :
+            corr = corr**2
+        distm = 1 - corr
+    else :
+        from scipy.spatial.distance import pdist,squareform
+        distm = squareform( pdist( crds , metric = distance_type ))
+    if bRemoveCurse : # EXPERIMENTAL
+        from impetuous.reducer import remove_curse
+        distm = remove_curse ( distm , nRound = nRound )
+    return ( distm )
+
+
 def confusion_matrix_df ( dict_row:dict , dict_col:dict , bSwitchKeyValues:str=True, bCheck:str=False ) -> pd.DataFrame :
     if bSwitchKeyValues :
         dict_row = invert_dict(dict_row)
@@ -1874,7 +1899,7 @@ def multivariate_aligned_pca ( analytes_df , journal_df ,
     corr_r = rankdata(pcas_df.T.apply(lambda x:np.sum(x**2)).values)/len(pcas_df.index)
 
     pcaw_df .loc[:,what] = [ dict_actual[s] for s in pcaw_df.index.values ]
-    projection_df       = pcaw_df.groupby(what).apply(np.mean)
+    projection_df       = pcaw_df.groupby(what).mean() #.apply(np.mean)
     projection_df       = ( projection_df.T / np.sqrt(np.sum(projection_df.values**2,1)) ).T
     projected_df        = pd.DataFrame( np.dot(projection_df,pcas_df.T), index=projection_df.index, columns=pcas_df.index )
     owners  = projected_df.index.values[projected_df.apply(np.abs).apply(np.argmax).values]
@@ -1906,6 +1931,7 @@ def sort_matrix ( matrix:np.array , linkage:str='single' ) -> list[int] :
     if linkage != 'dev' :
         from scipy.cluster import hierarchy
         from scipy.spatial.distance import squareform
+        from impetuous.clustering import absolute_coordinates_to_distance_matrix
         distm   = absolute_coordinates_to_distance_matrix ( matrix )
         pdi     = squareform ( distm )
         Z       = hierarchy.linkage( pdi , linkage )
@@ -1955,7 +1981,7 @@ def blind_confusion ( v1:list[str] , v2:list[str] , sort_type:str=None , bAbsolu
     else :
         return ( [ MAT,i_order,j_order ] )
 
-def confusion_matrix ( dict_row:dict , dict_col:dict , bSwitchKeyValues:bool=True ) -> dict :
+def confusion_matrix ( dict_row:dict , dict_col:dict , bSwitchKeyValues:bool=False ) -> dict :
     if bSwitchKeyValues :
         dict_row = invert_dict(dict_row)
         dict_col = invert_dict(dict_col)
