@@ -1,5 +1,5 @@
 """
-Copyright 2021 RICHARD TJÖRNHAMMAR
+Copyright 2022 RICHARD TJÖRNHAMMAR
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -479,6 +479,67 @@ def hyper_rdf ( df_ , label = 'generic' , sep = ',' , power=1. ,
     else :
         return ( rdf_ )
 
+def inplace_dimred ( x , n=2 , bSamples=True , dimr_type = 'PCA' ) :
+    rownames    = x.index.values
+    colnames    = x.columns.values
+    names       = rownames
+
+    if dimr_type == 'PCA' :
+        from sklearn.decomposition import PCA
+        ndims       = 2
+        pca         = PCA ( ndims )
+        if bSamples :
+            in_crds = x.iloc[:,:-1].T.values
+        else :
+            in_crds = x.iloc[:,:-1].values
+            names = colnames
+            colnames = rownames
+            rownames = names
+
+        crds = pca.fit_transform ( in_crds )
+        crds = pca.components_.T
+
+    if dimr_type == 'UMAP' :
+        import umap
+        umappad = umap.UMAP()
+        if bSamples :
+            crds    = umappad.fit_transform( x.values )
+        else :
+            crds    = umappad.fit_transform( x.T.values )
+            names = colnames
+            colnames = rownames
+            rownames = names
+
+    if True :
+        glab = x.iloc[0,-1]
+        colnames = [ *[ 'C'+str(i) for i in range(n) ], *['Name']]
+        book = pd.concat( [ pd.DataFrame( crds ).T , pd.DataFrame( [names] ) ] ).T
+        book.columns = colnames
+        book.index   = rownames
+    return ( book.dropna() )
+
+def remove_curse(cursed_distm:np.array, nRound:int=None ) -> np.array :
+    """SAIGA SPEAKING:
+REMOVE THE CURSE OF DIMENSIONALITY FROM
+A CURSED DISTANCE MATRIX BY PERFORMING A
+HISTOGRAM EQUALISATION ON ALL THE DISTANCES
+USING AVERAGE RANKING. THE DECIMAL ACCURACY
+MIGHT BE COMPROMISED IN THE UNCURSED DISTANCE
+MATRIX SO WE YIELD A DECIMAL PRUNED ONE IF
+REQUESTED"""
+    bounds = lambda rv:tuple((np.min(rv),np.max(rv)))
+    from scipy.stats import rankdata
+    n,m  = np.shape(cursed_distm)
+    rd_  = rankdata( cursed_distm.reshape(-1), 'average' )
+    dmin,dmax = bounds( list(set(rd_)) )
+    uncursed = (rd_/dmax-dmin/dmax).reshape(n,m)
+    if not nRound is None :
+        decimal_power = 10**(int(np.log10(n)) + nRound )
+        uncursed = np.array( [ np.round(v*decimal_power)/decimal_power for v in uncursed.reshape(-1)] ).reshape(n,m)
+        uncursed = uncursed*(1-np.eye(n))
+    return(uncursed)
+
+
 
 if __name__ == '__main__' :
     #
